@@ -1,6 +1,6 @@
 var LibraryPThread = {
   $PThread__postset: 'if (!ENVIRONMENT_IS_PTHREAD) PThread.initMainThreadBlock();',
-  $PThread__deps: ['$PROCINFO'],
+  $PThread__deps: ['$PROCINFO', '_register_pthread_ptr'],
   $PThread: {
     MAIN_THREAD_ID: 1, // A special constant that identifies the main JS thread ID.
     mainThreadInfo: {
@@ -17,6 +17,8 @@ var LibraryPThread = {
     initMainThreadBlock: function() {
       if (ENVIRONMENT_IS_PTHREAD) return undefined;
       PThread.mainThreadBlock = allocate({{{ C_STRUCTS.pthread.__size__ }}}, "i32*", ALLOC_STATIC);
+      __register_pthread_ptr(PThread.mainThreadBlock, /*isMainBrowserThread=*/!ENVIRONMENT_IS_WORKER, /*isMainRuntimeThread=*/1); // Pass the thread address inside the asm.js scope to store it for fast access that avoids the need for a FFI out.
+
       for (var i = 0; i < {{{ C_STRUCTS.pthread.__size__ }}}/4; ++i) HEAPU32[PThread.mainThreadBlock/4+i] = 0;
 
       // The pthread struct has a field that points to itself - this is used as a magic ID to detect whether the pthread_t
@@ -144,6 +146,7 @@ var LibraryPThread = {
         PThread.runExitHandlers();
 
         _emscripten_futex_wake(tb + {{{ C_STRUCTS.pthread.threadStatus }}}, {{{ cDefine('INT_MAX') }}});
+        __register_pthread_ptr(0, 0, 0); // Unregister the thread block also inside the asm.js scope.
         threadInfoStruct = 0;
         if (ENVIRONMENT_IS_PTHREAD) {
           postMessage({ cmd: 'exit' });
@@ -157,6 +160,7 @@ var LibraryPThread = {
       Atomics.store(HEAPU32, (threadInfoStruct + {{{ C_STRUCTS.pthread.threadStatus }}} ) >> 2, 1); // Mark the thread as no longer running.
       _emscripten_futex_wake(threadInfoStruct + {{{ C_STRUCTS.pthread.threadStatus }}}, {{{ cDefine('INT_MAX') }}}); // wake all threads
       threadInfoStruct = selfThreadId = 0; // Not hosting a pthread anymore in this worker, reset the info structures to null.
+      __register_pthread_ptr(0, 0, 0); // Unregister the thread block also inside the asm.js scope.
       postMessage({ cmd: 'cancelDone' });
     },
 
@@ -607,18 +611,38 @@ var LibraryPThread = {
     else PThread.threadExit(status);
   },
 
+  _pthread_ptr: 0,
+  _pthread_is_main_runtime_thread: 0,
+  _pthread_is_main_browser_thread: 0,
+
+  _register_pthread_ptr__asm: true,
+  _register_pthread_ptr__deps: ['_pthread_ptr', '_pthread_is_main_runtime_thread', '_pthread_is_main_browser_thread'],
+  _register_pthread_ptr: function(pthreadPtr, isMainBrowserThread, isMainRuntimeThread) {
+    pthreadPtr = pthreadPtr|0;
+    isMainBrowserThread = isMainBrowserThread|0;
+    isMainRuntimeThread = isMainRuntimeThread|0;
+    __pthread_ptr = pthreadPtr;
+    __pthread_is_main_browser_thread = isMainBrowserThread;
+    __pthread_is_main_runtime_thread = isMainRuntimeThread;
+  },
+
   // Public pthread_self() function which returns a unique ID for the thread.
+  pthread_self__asm: true,
+  pthread_self__deps: ['_pthread_ptr'],
   pthread_self: function() {
-    if (ENVIRONMENT_IS_PTHREAD) return threadInfoStruct;
-    return PThread.mainThreadBlock; // Main JS thread.
+    return __pthread_ptr|0;
   },
 
+  emscripten_is_main_runtime_thread__asm: true,
+  emscripten_is_main_runtime_thread__deps: ['_pthread_is_main_runtime_thread'],
   emscripten_is_main_runtime_thread: function() {
-    return !ENVIRONMENT_IS_PTHREAD;
+    return __pthread_is_main_runtime_thread|0; // Semantically the same as testing "!ENVIRONMENT_IS_PTHREAD" outside the asm.js scope
   },
 
+  emscripten_is_main_browser_thread__asm: true,
+  emscripten_is_main_browser_thread__deps: ['_pthread_is_main_browser_thread'],
   emscripten_is_main_browser_thread: function() {
-    return !ENVIRONMENT_IS_WORKER;
+    return __pthread_is_main_browser_thread|0; // Semantically the same as testing "!ENVIRONMENT_IS_WORKER" outside the asm.js scope
   },
 
   pthread_getschedparam: function(thread, policy, schedparam) {
@@ -743,43 +767,118 @@ var LibraryPThread = {
     return 0;
   },
 
+  // Stores the memory address that the main thread is futexWaiting on, if any.
+  _main_thread_futex_wait_address: '; if (ENVIRONMENT_IS_PTHREAD) __main_thread_futex_wait_address = PthreadWorkerInit.__main_thread_futex_wait_address; else PthreadWorkerInit.__main_thread_futex_wait_address = __main_thread_futex_wait_address = allocate(1, "i32*", ALLOC_STATIC)',
+
   // Returns 0 on success, or one of the values -ETIMEDOUT, -EWOULDBLOCK or -EINVAL on error.
+  emscripten_futex_wait__deps: ['_main_thread_futex_wait_address'],
   emscripten_futex_wait: function(addr, val, timeout) {
     if (addr <= 0 || addr > HEAP8.length || addr&3 != 0) return -{{{ cDefine('EINVAL') }}};
 //    dump('futex_wait addr:' + addr + ' by thread: ' + _pthread_self() + (ENVIRONMENT_IS_PTHREAD?'(pthread)':'') + '\n');
+    if (ENVIRONMENT_IS_WORKER) {
 #if PTHREADS_PROFILING
-    PThread.setThreadStatusConditional(_pthread_self(), {{{ cDefine('EM_THREAD_STATUS_RUNNING') }}}, {{{ cDefine('EM_THREAD_STATUS_WAITFUTEX') }}});
+      PThread.setThreadStatusConditional(_pthread_self(), {{{ cDefine('EM_THREAD_STATUS_RUNNING') }}}, {{{ cDefine('EM_THREAD_STATUS_WAITFUTEX') }}});
 #endif
-    var ret = Atomics.futexWait(HEAP32, addr >> 2, val, timeout);
+      var ret = Atomics.futexWait(HEAP32, addr >> 2, val, timeout);
 //    dump('futex_wait done by thread: ' + _pthread_self() + (ENVIRONMENT_IS_PTHREAD?'(pthread)':'') + '\n');
 #if PTHREADS_PROFILING
-    PThread.setThreadStatusConditional(_pthread_self(), {{{ cDefine('EM_THREAD_STATUS_WAITFUTEX') }}}, {{{ cDefine('EM_THREAD_STATUS_RUNNING') }}});
+      PThread.setThreadStatusConditional(_pthread_self(), {{{ cDefine('EM_THREAD_STATUS_WAITFUTEX') }}}, {{{ cDefine('EM_THREAD_STATUS_RUNNING') }}});
 #endif
-    if (ret == Atomics.TIMEDOUT) return -{{{ cDefine('ETIMEDOUT') }}};
-    if (ret == Atomics.NOTEQUAL) return -{{{ cDefine('EWOULDBLOCK') }}};
-    if (ret == 0) return 0;
-    throw 'Atomics.futexWait returned an unexpected value ' + ret;
+      if (ret == Atomics.TIMEDOUT) return -{{{ cDefine('ETIMEDOUT') }}};
+      if (ret == Atomics.NOTEQUAL) return -{{{ cDefine('EWOULDBLOCK') }}};
+      if (ret == 0) return 0;
+      throw 'Atomics.futexWait returned an unexpected value ' + ret;
+    } else {
+      // Atomics.futexWait is not available in the main browser thread, so simulate it via busy spinning.
+      var loadedVal = Atomics.load(HEAP32, addr >> 2);
+      if (val != loadedVal) return -{{{ cDefine('EWOULDBLOCK') }}};
+
+      var tNow = performance.now();
+      var tEnd = tNow + timeout;
+
+#if PTHREADS_PROFILING
+      PThread.setThreadStatusConditional(_pthread_self(), {{{ cDefine('EM_THREAD_STATUS_RUNNING') }}}, {{{ cDefine('EM_THREAD_STATUS_WAITFUTEX') }}});
+#endif
+
+      // Register globally which address the main thread is simulating to be futexWaiting on. When zero, main thread is not waiting on anything,
+      // and on nonzero, the contents of address pointed by __main_thread_futex_wait_address tell which address the main thread is simulating its wait on.
+      Atomics.store(HEAP32, __main_thread_futex_wait_address >> 2, addr);
+      while (addr) {
+        tNow = performance.now();
+        if (tNow > tEnd) {
+#if PTHREADS_PROFILING
+          PThread.setThreadStatusConditional(_pthread_self(), {{{ cDefine('EM_THREAD_STATUS_RUNNING') }}}, {{{ cDefine('EM_THREAD_STATUS_WAITFUTEX') }}});
+#endif
+          return -{{{ cDefine('ETIMEDOUT') }}};
+        }
+        addr = Atomics.load(HEAP32, __main_thread_futex_wait_address >> 2); // Look for a worker thread waking us up.
+      }
+#if PTHREADS_PROFILING
+      PThread.setThreadStatusConditional(_pthread_self(), {{{ cDefine('EM_THREAD_STATUS_RUNNING') }}}, {{{ cDefine('EM_THREAD_STATUS_WAITFUTEX') }}});
+#endif
+      return 0;
+    }
   },
 
   // Returns the number of threads (>= 0) woken up, or the value -EINVAL on error.
   // Pass count == INT_MAX to wake up all threads.
+  emscripten_futex_wake__deps: ['_main_thread_futex_wait_address'],
   emscripten_futex_wake: function(addr, count) {
     if (addr <= 0 || addr > HEAP8.length || addr&3 != 0 || count < 0) return -{{{ cDefine('EINVAL') }}};
+    if (count == 0) return 0;
 //    dump('futex_wake addr:' + addr + ' by thread: ' + _pthread_self() + (ENVIRONMENT_IS_PTHREAD?'(pthread)':'') + '\n');
+
+    // See if main thread is waiting on this address? If so, wake it up by resetting its wake location to zero.
+    // Note that this is not a fair procedure, since we always wake main thread first before any workers, so
+    // this scheme does not adhere to real queue-based waiting.
+    var mainThreadWaitAddress = Atomics.load(HEAP32, __main_thread_futex_wait_address >> 2);
+    var mainThreadWoken = 0;
+    if (mainThreadWaitAddress == addr) {
+      var loadedAddr = Atomics.compareExchange(HEAP32, __main_thread_futex_wait_address >> 2, mainThreadWaitAddress, 0);
+      if (loadedAddr == mainThreadWaitAddress) {
+        --count;
+        mainThreadWoken = 1;
+        if (count <= 0) return 1;
+      }
+    }
+
+    // Wake any workers waiting on this address.
     var ret = Atomics.futexWake(HEAP32, addr >> 2, count);
-    if (ret >= 0) return ret;
+    if (ret >= 0) return ret + mainThreadWoken;
     throw 'Atomics.futexWake returned an unexpected value ' + ret;
   },
 
   // Returns the number of threads (>= 0) woken up, or one of the values -EINVAL or -EAGAIN on error.
+  emscripten_futex_wake_or_requeue__deps: ['_main_thread_futex_wait_address'],
   emscripten_futex_wake_or_requeue: function(addr, count, cmpValue, addr2) {
     if (addr <= 0 || addr2 <= 0 || addr >= HEAP8.length || addr2 >= HEAP8.length || count < 0
       || addr&3 != 0 || addr2&3 != 0) {
       return -{{{ cDefine('EINVAL') }}};
     }
+
+    // See if main thread is waiting on this address? If so, wake it up by resetting its wake location to zero,
+    // or move it to wait on addr2. Note that this is not a fair procedure, since we always wake main thread first before
+    // any workers, so this scheme does not adhere to real queue-based waiting.
+    var mainThreadWaitAddress = Atomics.load(HEAP32, __main_thread_futex_wait_address >> 2);
+    var mainThreadWoken = 0;
+    if (mainThreadWaitAddress == addr) {
+      // Check cmpValue precondition before taking any action.
+      var val1 = Atomics.load(HEAP32, addr >> 2);
+      if (val1 != cmpValue) return -{{{ cDefine('EAGAIN') }}};
+
+      // If we are actually waking any waiters, then new main thread wait location is reset, otherwise requeue it to wait on addr2.
+      var newMainThreadWaitAddress = (count > 0) ? 0 : addr2;
+      var loadedAddr = Atomics.compareExchange(HEAP32, __main_thread_futex_wait_address >> 2, mainThreadWaitAddress, newMainThreadWaitAddress);
+      if (loadedAddr == mainThreadWaitAddress && count > 0) {
+        --count; // Main thread was woken, so one less workers to wake up.
+        mainThreadWoken = 1;
+      }
+    }
+
+    // Wake any workers waiting on this address.
     var ret = Atomics.futexWakeOrRequeue(HEAP32, addr >> 2, count, cmpValue, addr >> 2);
     if (ret == Atomics.NOTEQUAL) return -{{{ cDefine('EAGAIN') }}};
-    if (ret >= 0) return ret;
+    if (ret >= 0) return ret + mainThreadWoken;
     throw 'Atomics.futexWakeOrRequeue returned an unexpected value ' + ret;
   },
 
